# Copyright (C) 2016-present the asyncpg authors and contributors
# <see AUTHORS file>
#
# This module is part of asyncpg and is released under
# the Apache 2.0 License: http://www.apache.org/licenses/LICENSE-2.0


cimport cython
cimport cpython

from libc.stdint cimport int16_t, int32_t, uint16_t, uint32_t, int64_t, uint64_t


include "./consts.pxi"
include "./frb.pxd"
include "./buffer.pxd"


include "./codecs/__init__.pxd"



# Copyright (C) 2016-present the asyncpg authors and contributors
# <see AUTHORS file>
#
# This module is part of asyncpg and is released under
# the Apache 2.0 License: http://www.apache.org/licenses/LICENSE-2.0


cdef bits_encode(CodecContext settings, WriteBuffer wbuf, obj):
    cdef:
        Py_buffer pybuf
        bint pybuf_used = False
        char *buf
        ssize_t len
        ssize_t bitlen

    if cpython.PyBytes_CheckExact(obj):
        buf = cpython.PyBytes_AS_STRING(obj)
        len = cpython.Py_SIZE(obj)
        bitlen = len * 8
    elif isinstance(obj, pgproto_types.BitString):
        cpython.PyBytes_AsStringAndSize(obj.bytes, &buf, &len)
        bitlen = obj.__len__()
    else:
        cpython.PyObject_GetBuffer(obj, &pybuf, cpython.PyBUF_SIMPLE)
        pybuf_used = True
        buf = <char*>pybuf.buf
        len = pybuf.len
        bitlen = len * 8

    try:
        if bitlen > _MAXINT32:
            raise ValueError('bit value too long')
        wbuf.write_int32(4 + <int32_t>len)
        wbuf.write_int32(<int32_t>bitlen)
        wbuf.write_cstr(buf, len)
    finally:
        if pybuf_used:
            cpython.PyBuffer_Release(&pybuf)


cdef bits_decode(CodecContext settings, FRBuffer *buf):
    cdef:
        int32_t bitlen = hton.unpack_int32(frb_read(buf, 4))
        ssize_t buf_len = buf.len

    bytes_ = cpython.PyBytes_FromStringAndSize(frb_read_all(buf), buf_len)
    return pgproto_types.BitString.frombytes(bytes_, bitlen)



# Copyright (C) 2016-present the asyncpg authors and contributors
# <see AUTHORS file>
#
# This module is part of asyncpg and is released under
# the Apache 2.0 License: http://www.apache.org/licenses/LICENSE-2.0


from asyncpg import exceptions


cdef inline record_encode_frame(ConnectionSettings settings, WriteBuffer buf,
                                WriteBuffer elem_data, int32_t elem_count):
    buf.write_int32(4 + elem_data.len())
    # attribute count
    buf.write_int32(elem_count)
    # encoded attribute data
    buf.write_buffer(elem_data)


cdef anonymous_record_decode(ConnectionSettings settings, FRBuffer *buf):
    cdef:
        tuple result
        ssize_t elem_count
        ssize_t i
        int32_t elem_len
        uint32_t elem_typ
        Codec elem_codec
        FRBuffer elem_buf

    elem_count = <ssize_t><uint32_t>hton.unpack_int32(frb_read(buf, 4))
    result = cpython.PyTuple_New(elem_count)

    for i in range(elem_count):
        elem_typ = <uint32_t>hton.unpack_int32(frb_read(buf, 4))
        elem_len = hton.unpack_int32(frb_read(buf, 4))

        if elem_len == -1:
            elem = None
        else:
            elem_codec = settings.get_data_codec(elem_typ)
            if elem_codec is None or not elem_codec.has_decoder():
                raise exceptions.InternalClientError(
                    'no decoder for composite type element in '
                    'position {} of type OID {}'.format(i, elem_typ))
            elem = elem_codec.decode(settings,
                                     frb_slice_from(&elem_buf, buf, elem_len))

        cpython.Py_INCREF(elem)
        cpython.PyTuple_SET_ITEM(result, i, elem)

    return result


cdef anonymous_record_encode(ConnectionSettings settings, WriteBuffer buf, obj):
    raise exceptions.UnsupportedClientFeatureError(
        'input of anonymous composite types is not supported',
        hint=(
            'Consider declaring an explicit composite type and '
            'using it to cast the argument.'
        ),
        detail='PostgreSQL does not implement anonymous composite type input.'
    )


cdef init_record_codecs():
    register_core_codec(RECORDOID,
                        <encode_func>anonymous_record_encode,
                        <decode_func>anonymous_record_decode,
                        PG_FORMAT_BINARY)

init_record_codecs()



# Copyright (C) 2016-present the asyncpg authors and contributors
# <see AUTHORS file>
#
# This module is part of asyncpg and is released under
# the Apache 2.0 License: http://www.apache.org/licenses/LICENSE-2.0


# GENERATED FROM pg_catalog.pg_type
# DO NOT MODIFY, use tools/generate_type_map.py to update

DEF INVALIDOID = 0
DEF MAXBUILTINOID = 9999
DEF MAXSUPPORTEDOID = 5080

DEF BOOLOID = 16
DEF BYTEAOID = 17
DEF CHAROID = 18
DEF NAMEOID = 19
DEF INT8OID = 20
DEF INT2OID = 21
DEF INT4OID = 23
DEF REGPROCOID = 24
DEF TEXTOID = 25
DEF OIDOID = 26
DEF TIDOID = 27
DEF XIDOID = 28
DEF CIDOID = 29
DEF PG_DDL_COMMANDOID = 32
DEF JSONOID = 114
DEF XMLOID = 142
DEF PG_NODE_TREEOID = 194
DEF SMGROID = 210
DEF TABLE_AM_HANDLEROID = 269
DEF INDEX_AM_HANDLEROID = 325
DEF POINTOID = 600
DEF LSEGOID = 601
DEF PATHOID = 602
DEF BOXOID = 603
DEF POLYGONOID = 604
DEF LINEOID = 628
DEF CIDROID = 650
DEF FLOAT4OID = 700
DEF FLOAT8OID = 701
DEF ABSTIMEOID = 702
DEF RELTIMEOID = 703
DEF TINTERVALOID = 704
DEF UNKNOWNOID = 705
DEF CIRCLEOID = 718
DEF MACADDR8OID = 774
DEF MONEYOID = 790
DEF MACADDROID = 829
DEF INETOID = 869
DEF _TEXTOID = 1009
DEF _OIDOID = 1028
DEF ACLITEMOID = 1033
DEF BPCHAROID = 1042
DEF VARCHAROID = 1043
DEF DATEOID = 1082
DEF TIMEOID = 1083
DEF TIMESTAMPOID = 1114
DEF TIMESTAMPTZOID = 1184
DEF INTERVALOID = 1186
DEF TIMETZOID = 1266
DEF BITOID = 1560
DEF VARBITOID = 1562
DEF NUMERICOID = 1700
DEF REFCURSOROID = 1790
DEF REGPROCEDUREOID = 2202
DEF REGOPEROID = 2203
DEF REGOPERATOROID = 2204
DEF REGCLASSOID = 2205
DEF REGTYPEOID = 2206
DEF RECORDOID = 2249
DEF CSTRINGOID = 2275
DEF ANYOID = 2276
DEF ANYARRAYOID = 2277
DEF VOIDOID = 2278
DEF TRIGGEROID = 2279
DEF LANGUAGE_HANDLEROID = 2280
DEF INTERNALOID = 2281
DEF OPAQUEOID = 2282
DEF ANYELEMENTOID = 2283
DEF ANYNONARRAYOID = 2776
DEF UUIDOID = 2950
DEF TXID_SNAPSHOTOID = 2970
DEF FDW_HANDLEROID = 3115
DEF PG_LSNOID = 3220
DEF TSM_HANDLEROID = 3310
DEF PG_NDISTINCTOID = 3361
DEF PG_DEPENDENCIESOID = 3402
DEF ANYENUMOID = 3500
DEF TSVECTOROID = 3614
DEF TSQUERYOID = 3615
DEF GTSVECTOROID = 3642
DEF REGCONFIGOID = 3734
DEF REGDICTIONARYOID = 3769
DEF JSONBOID = 3802
DEF ANYRANGEOID = 3831
DEF EVENT_TRIGGEROID = 3838
DEF JSONPATHOID = 4072
DEF REGNAMESPACEOID = 4089
DEF REGROLEOID = 4096
DEF REGCOLLATIONOID = 4191
DEF PG_MCV_LISTOID = 5017
DEF PG_SNAPSHOTOID = 5038
DEF XID8OID = 5069
DEF ANYCOMPATIBLEOID = 5077
DEF ANYCOMPATIBLEARRAYOID = 5078
DEF ANYCOMPATIBLENONARRAYOID = 5079
DEF ANYCOMPATIBLERANGEOID = 5080

cdef ARRAY_TYPES = (_TEXTOID, _OIDOID,)

BUILTIN_TYPE_OID_MAP = {
    ABSTIMEOID: 'abstime',
    ACLITEMOID: 'aclitem',
    ANYARRAYOID: 'anyarray',
    ANYCOMPATIBLEARRAYOID: 'anycompatiblearray',
    ANYCOMPATIBLENONARRAYOID: 'anycompatiblenonarray',
    ANYCOMPATIBLEOID: 'anycompatible',
    ANYCOMPATIBLERANGEOID: 'anycompatiblerange',
    ANYELEMENTOID: 'anyelement',
    ANYENUMOID: 'anyenum',
    ANYNONARRAYOID: 'anynonarray',
    ANYOID: 'any',
    ANYRANGEOID: 'anyrange',
    BITOID: 'bit',
    BOOLOID: 'bool',
    BOXOID: 'box',
    BPCHAROID: 'bpchar',
    BYTEAOID: 'bytea',
    CHAROID: 'char',
    CIDOID: 'cid',
    CIDROID: 'cidr',
    CIRCLEOID: 'circle',
    CSTRINGOID: 'cstring',
    DATEOID: 'date',
    EVENT_TRIGGEROID: 'event_trigger',
    FDW_HANDLEROID: 'fdw_handler',
    FLOAT4OID: 'float4',
    FLOAT8OID: 'float8',
    GTSVECTOROID: 'gtsvector',
    INDEX_AM_HANDLEROID: 'index_am_handler',
    INETOID: 'inet',
    INT2OID: 'int2',
    INT4OID: 'int4',
    INT8OID: 'int8',
    INTERNALOID: 'internal',
    INTERVALOID: 'interval',
    JSONBOID: 'jsonb',
    JSONOID: 'json',
    JSONPATHOID: 'jsonpath',
    LANGUAGE_HANDLEROID: 'language_handler',
    LINEOID: 'line',
    LSEGOID: 'lseg',
    MACADDR8OID: 'macaddr8',
    MACADDROID: 'macaddr',
    MONEYOID: 'money',
    NAMEOID: 'name',
    NUMERICOID: 'numeric',
    OIDOID: 'oid',
    OPAQUEOID: 'opaque',
    PATHOID: 'path',
    PG_DDL_COMMANDOID: 'pg_ddl_command',
    PG_DEPENDENCIESOID: 'pg_dependencies',
    PG_LSNOID: 'pg_lsn',
    PG_MCV_LISTOID: 'pg_mcv_list',
    PG_NDISTINCTOID: 'pg_ndistinct',
    PG_NODE_TREEOID: 'pg_node_tree',
    PG_SNAPSHOTOID: 'pg_snapshot',
    POINTOID: 'point',
    POLYGONOID: 'polygon',
    RECORDOID: 'record',
    REFCURSOROID: 'refcursor',
    REGCLASSOID: 'regclass',
    REGCOLLATIONOID: 'regcollation',
    REGCONFIGOID: 'regconfig',
    REGDICTIONARYOID: 'regdictionary',
    REGNAMESPACEOID: 'regnamespace',
    REGOPERATOROID: 'regoperator',
    REGOPEROID: 'regoper',
    REGPROCEDUREOID: 'regprocedure',
    REGPROCOID: 'regproc',
    REGROLEOID: 'regrole',
    REGTYPEOID: 'regtype',
    RELTIMEOID: 'reltime',
    SMGROID: 'smgr',
    TABLE_AM_HANDLEROID: 'table_am_handler',
    TEXTOID: 'text',
    TIDOID: 'tid',
    TIMEOID: 'time',
    TIMESTAMPOID: 'timestamp',
    TIMESTAMPTZOID: 'timestamptz',
    TIMETZOID: 'timetz',
    TINTERVALOID: 'tinterval',
    TRIGGEROID: 'trigger',
    TSM_HANDLEROID: 'tsm_handler',
    TSQUERYOID: 'tsquery',
    TSVECTOROID: 'tsvector',
    TXID_SNAPSHOTOID: 'txid_snapshot',
    UNKNOWNOID: 'unknown',
    UUIDOID: 'uuid',
    VARBITOID: 'varbit',
    VARCHAROID: 'varchar',
    VOIDOID: 'void',
    XID8OID: 'xid8',
    XIDOID: 'xid',
    XMLOID: 'xml',
    _OIDOID: 'oid[]',
    _TEXTOID: 'text[]'
}

BUILTIN_TYPE_NAME_MAP = {v: k for k, v in BUILTIN_TYPE_OID_MAP.items()}

BUILTIN_TYPE_NAME_MAP['smallint'] = \
    BUILTIN_TYPE_NAME_MAP['int2']

BUILTIN_TYPE_NAME_MAP['int'] = \
    BUILTIN_TYPE_NAME_MAP['int4']

BUILTIN_TYPE_NAME_MAP['integer'] = \
    BUILTIN_TYPE_NAME_MAP['int4']

BUILTIN_TYPE_NAME_MAP['bigint'] = \
    BUILTIN_TYPE_NAME_MAP['int8']

BUILTIN_TYPE_NAME_MAP['decimal'] = \
    BUILTIN_TYPE_NAME_MAP['numeric']

BUILTIN_TYPE_NAME_MAP['real'] = \
    BUILTIN_TYPE_NAME_MAP['float4']

BUILTIN_TYPE_NAME_MAP['double precision'] = \
    BUILTIN_TYPE_NAME_MAP['float8']

BUILTIN_TYPE_NAME_MAP['timestamp with timezone'] = \
    BUILTIN_TYPE_NAME_MAP['timestamptz']

BUILTIN_TYPE_NAME_MAP['timestamp without timezone'] = \
    BUILTIN_TYPE_NAME_MAP['timestamp']

BUILTIN_TYPE_NAME_MAP['time with timezone'] = \
    BUILTIN_TYPE_NAME_MAP['timetz']

BUILTIN_TYPE_NAME_MAP['time without timezone'] = \
    BUILTIN_TYPE_NAME_MAP['time']

BUILTIN_TYPE_NAME_MAP['char'] = \
    BUILTIN_TYPE_NAME_MAP['bpchar']

BUILTIN_TYPE_NAME_MAP['character'] = \
    BUILTIN_TYPE_NAME_MAP['bpchar']

BUILTIN_TYPE_NAME_MAP['character varying'] = \
    BUILTIN_TYPE_NAME_MAP['varchar']

BUILTIN_TYPE_NAME_MAP['bit varying'] = \
    BUILTIN_TYPE_NAME_MAP['varbit']



# Copyright (C) 2016-present the asyncpg authors and contributors
# <see AUTHORS file>
#
# This module is part of asyncpg and is released under
# the Apache 2.0 License: http://www.apache.org/licenses/LICENSE-2.0


import ipaddress


# defined in postgresql/src/include/inet.h
#
DEF PGSQL_AF_INET = 2  # AF_INET
DEF PGSQL_AF_INET6 = 3  # AF_INET + 1


_ipaddr = ipaddress.ip_address
_ipiface = ipaddress.ip_interface
_ipnet = ipaddress.ip_network


cdef inline uint8_t _ip_max_prefix_len(int32_t family):
    # Maximum number of bits in the network prefix of the specified
    # IP protocol version.
    if family == PGSQL_AF_INET:
        return 32
    else:
        return 128


cdef inline int32_t _ip_addr_len(int32_t family):
    # Length of address in bytes for the specified IP protocol version.
    if family == PGSQL_AF_INET:
        return 4
    else:
        return 16


cdef inline int8_t _ver_to_family(int32_t version):
    if version == 4:
        return PGSQL_AF_INET
    else:
        return PGSQL_AF_INET6


cdef inline _net_encode(WriteBuffer buf, int8_t family, uint32_t bits,
                        int8_t is_cidr, bytes addr):

    cdef:
        char *addrbytes
        ssize_t addrlen

    cpython.PyBytes_AsStringAndSize(addr, &addrbytes, &addrlen)

    buf.write_int32(4 + <int32_t>addrlen)
    buf.write_byte(family)
    buf.write_byte(<int8_t>bits)
    buf.write_byte(is_cidr)
    buf.write_byte(<int8_t>addrlen)
    buf.write_cstr(addrbytes, addrlen)


cdef net_decode(CodecContext settings, FRBuffer *buf, bint as_cidr):
    cdef:
        int32_t family = <int32_t>frb_read(buf, 1)[0]
        uint8_t bits = <uint8_t>frb_read(buf, 1)[0]
        int prefix_len
        int32_t is_cidr = <int32_t>frb_read(buf, 1)[0]
        int32_t addrlen = <int32_t>frb_read(buf, 1)[0]
        bytes addr
        uint8_t max_prefix_len = _ip_max_prefix_len(family)

    if is_cidr != as_cidr:
        raise ValueError('unexpected CIDR flag set in non-cidr value')

    if family != PGSQL_AF_INET and family != PGSQL_AF_INET6:
        raise ValueError('invalid address family in "{}" value'.format(
            'cidr' if is_cidr else 'inet'
        ))

    max_prefix_len = _ip_max_prefix_len(family)

    if bits > max_prefix_len:
        raise ValueError('invalid network prefix length in "{}" value'.format(
            'cidr' if is_cidr else 'inet'
        ))

    if addrlen != _ip_addr_len(family):
        raise ValueError('invalid address length in "{}" value'.format(
            'cidr' if is_cidr else 'inet'
        ))

    addr = cpython.PyBytes_FromStringAndSize(frb_read(buf, addrlen), addrlen)

    if as_cidr or bits != max_prefix_len:
        prefix_len = cpython.PyLong_FromLong(bits)

        if as_cidr:
            return _ipnet((addr, prefix_len))
        else:
            return _ipiface((addr, prefix_len))
    else:
        return _ipaddr(addr)


cdef cidr_encode(CodecContext settings, WriteBuffer buf, obj):
    cdef:
        object ipnet
        int8_t family

    ipnet = _ipnet(obj)
    family = _ver_to_family(ipnet.version)
    _net_encode(buf, family, ipnet.prefixlen, 1, ipnet.network_address.packed)


cdef cidr_decode(CodecContext settings, FRBuffer *buf):
    return net_decode(settings, buf, True)


cdef inet_encode(CodecContext settings, WriteBuffer buf, obj):
    cdef:
        object ipaddr
        int8_t family

    try:
        ipaddr = _ipaddr(obj)
    except ValueError:
        # PostgreSQL accepts *both* CIDR and host values
        # for the host datatype.
        ipaddr = _ipiface(obj)
        family = _ver_to_family(ipaddr.version)
        _net_encode(buf, family, ipaddr.network.prefixlen, 1, ipaddr.packed)
    else:
        family = _ver_to_family(ipaddr.version)
        _net_encode(buf, family, _ip_max_prefix_len(family), 0, ipaddr.packed)


cdef inet_decode(CodecContext settings, FRBuffer *buf):
    return net_decode(settings, buf, False)
